---
title: An early Lisp scriptlang prototype in Ada
url: http://thetarpit.org/posts/y04/074-adalisp-prototype.markdown
---

This document describes the result of my attempts at implementing a Lisp
(more precisely a Scheme) scripting language in Ada. The effort spawned
from numerous discussions in the logs[^1] and used a few important
references[^2]. In short, the goal is to have a Lisp interpreter that is
a. small, easy to read, i.e. fits-in-head; and b. written in a sane
language, sans pointers and other cockroaches inhabiting today's
[broken computing machines][software-engineering].

I have attached below:

* a tarball: [adalisp-proto.tar.gz][adalisp-proto.tar.gz]; and
* a detached signature by yours truly:
  [adalisp-proto.tar.gz.spyked.sig][adalisp-proto.tar.gz.spyked.sig].

The source code is organized as follows:

* a set of restrictions for the implementation (`restrict.adc`)[^3];
* a Lisp machine representation in Ada (`lispm.{ads,adb}`);
* a S-expression parser (`parser.{ads,adb}`);
* a Scheme evaluator (`evaler.{ads,adb}`); and finally,
* some glue that puts everything together in a Read-Eval-Print Loop
  (`test_repl.adb`).

Additionally, the tarball includes a GPRbuild project file and a small
set of tests. To compile and run the tests, run in the project
directory:

~~~~
$ gprbuild
$ cat test.scm | ./bin/test_repl
~~~~

In the end, some comments about the implementation:

[i] The implementation is incomplete. A lot of functionality is still
unimplemented, e.g. `<`, `>`, `gensym`, quasiquotes, `let*`,
`letrec`. This is just a matter of adding them into the `evaler`.

[ii] The implementation is incomplete also because it has no mechanism
for recovery from parser and evaler errors, etc.

[iii] Reading from standard input is utterly broken in many ways,
because Ada's `Get_Char` doesn't know how to read newlines among
others. These should be replaced with FFA's [I/O routines][ffacalc-io].

[iv] The author's cockroaches led him to implement Scheme booleans
(`#t`, `#f`) as a subset of integers. These are traditionally
symbols. On another note, it's not immediately clear that schemisms such
as `#t`/`#f` are relevant at all, one might as well use `t`/`nil` as
truth values.

[vi] The code should be refactored in more than one direction. One
important direction is separating all the mechanisms in a `libadalisp`
to be used for embedding into other applications. Another direction
involves improving readability by e.g. using keyword arguments instead
of the positional ones that are so popular in C programs. But more
importantly, the user should be able to swallow the program piece by
piece, as is healthy and has been [demonstrated before][ffa-ch1].

[vii] I suspect the `evaler` is [far from correct][logs-evaluation], but
this can only be evaluated [sic!] after the completion of point vi
above.

[viii] The list is far from exhaustive. Some aspects are already being
discussed in the logs[^4]; other threads will inevitably arise.

[^1]: See for example: [1][logs-adalisp], [2][logs-ada-scheme],
    [3][logs-ada-lisp].

[^2]: Kogge's "The Architecture of Symbolic Computers", Queinnec's "Lisp
    in Small Pieces", (in some cases) the [tinyscheme][tinyscheme]
    implementation, the [FFA][ffa] implementation and last but not
    least, the [Ada reference manual][ada].

[^3]: For more details, consult the [FFA][ffa] series, in particular the
    [first chapter][ffa-ch1].

[^4]: [For example][logs-bytecode].

[logs-adalisp]: http://btcbase.org/log-search?q=adalisp
[logs-ada-scheme]: http://btcbase.org/log-search?q=ada+scheme
[logs-ada-lisp]: http://btcbase.org/log-search?q=ada+lisp
[tinyscheme]: http://tinyscheme.sourceforge.net/
[ffa]: http://www.loper-os.org/?cat=49
[ffa-ch1]: http://www.loper-os.org/?p=1913
[ada]: http://ada-auth.org/standards/ada12.html
[software-engineering]: http://thetarpit.org/posts/y02/03c-the-myth-of-software-engineering.html
[adalisp-proto.tar.gz]: http://thetarpit.org/uploads/2018/06/adalisp-proto.tar.gz
[adalisp-proto.tar.gz.spyked.sig]: http://thetarpit.org/uploads/2018/06/adalisp-proto.tar.gz.spyked.sig
[ffacalc-io]: http://btcbase.org/patches/ffa_ch4_ffacalc#L816
[logs-evaluation]: http://btcbase.org/log/2018-06-21#1828378
[logs-bytecode]: http://btcbase.org/log/2018-06-22#1828584
